from __future__ import unicode_literals, absolute_import

import logging
import re
from typing import Tuple

from aniparse import parser_helper
from aniparse.element import ElementCategory
from aniparse.keyword import KeywordManager
from aniparse.token import TokenCategory, Tokens

logger = logging.getLogger(__name__)

open_brackets = (
    "(",  # U+0028 LEFT PARENTHESIS
    "[",  # U+005B LEFT SQUARE BRACKET
    "{",  # U+007B LEFT CURLY BRACKET
    "\u300C",  # Corner bracket
    "\u300E",  # White corner bracket
    "\u3010",  # Black lenticular bracket
    "\uFF08",  # Fullwidth parenthesis
)

close_brackets = (
    ")",  # U+0029 Right parenthesis
    "]",  # U+005D Right square bracket
    "}",  # U+007D Right curly bracket
    "\u300D",  # Corner bracket
    "\u300F",  # White corner bracket
    "\u3011",  # Black lenticular bracket
    "\uFF09",  # Fullwidth right parenthesis
)

brackets = open_brackets + close_brackets
escaped_open_brackets = [re.escape(bracket) for bracket in open_brackets]
escaped_close_brackets = [re.escape(bracket) for bracket in close_brackets]
escaped_brackets = '|'.join(escaped_open_brackets + escaped_close_brackets)


def invert_string_brackets(text) -> Tuple[str, list, list]:
    close_brackets_position = []
    open_brackets_position = []
    reverted_name = ""
    for idx, char in enumerate(text[::-1]):
        if char in open_brackets:
            reverted_name += close_brackets[open_brackets.index(char)]
            close_brackets_position.append(idx)
        elif char in close_brackets:
            reverted_name += open_brackets[close_brackets.index(char)]
            open_brackets_position.append(idx)
        else:
            reverted_name += char
    return reverted_name, open_brackets_position, close_brackets_position


class Tokenizer(Tokens):
    def __init__(self, filename: str, options: dict, keyword_manager: KeywordManager):
        self.filename = filename
        self.options = options
        delimiters = ''.join(
            ['\\' + d for d in self.options['allowed_delimiters']])
        self.pattern = '([{0}])'.format(delimiters)
        super().__init__(keyword_manager)

    @staticmethod
    def _get_brackets_position(filename) -> Tuple[str, list, list]:
        # find all brackets positions from self.filename
        open_brackets_position = []
        close_brackets_position = []
        for bracket in re.finditer(escaped_brackets, filename, re.IGNORECASE):
            if bracket.group() in open_brackets:
                open_brackets_position.append(bracket.start())
            else:
                close_brackets_position.append(bracket.start())

        # if the length is not same, then there is a bracket that is not closed or typo
        if len(open_brackets_position) != len(close_brackets_position):
            # if the different is even, then it is probably a typo
            if (len(open_brackets_position) - len(close_brackets_position)) % 2 == 0:
                all_bracket = sorted(open_brackets_position + close_brackets_position)
                last_seen = []
                expected = []
                reverted = False
                if len(close_brackets_position) > len(open_brackets_position):
                    # reverse the filename and invert open brackets and close brackets
                    reverted = True
                    filename, open_brackets_position, close_brackets_position = invert_string_brackets(
                        filename)
                    all_bracket = sorted(open_brackets_position + close_brackets_position)

                for index in all_bracket:
                    char = filename[index]
                    if char in open_brackets:
                        try:
                            abc = char == last_seen[-1]
                        except IndexError:
                            abc = False
                        if abc:
                            # We expected a close bracket, but we got the same open bracket
                            # This is probably a typo, replace it with correct bracket
                            corrected = close_brackets[open_brackets.index(char)]
                            filename = filename[:index] + corrected + filename[index + 1:]
                            # remove this index from open_brackets_position and add to close_brackets_position
                            open_brackets_position.remove(index)
                            close_brackets_position.append(index)
                            last_seen.pop()
                        else:
                            last_seen.append(char)
                            expected.append(close_brackets[open_brackets.index(char)])

                    else:
                        try:
                            if char == expected[-1]:
                                last_seen.pop()
                                expected.pop()
                        except IndexError:
                            pass
                if reverted:
                    filename, open_brackets_position, close_brackets_position = invert_string_brackets(
                        filename)

            # else: I'm not sure what to do here. Let's just ignore it and put on log
            else:
                logger.debug(f"Invalid brackets in {filename}")
                # if the length not same, if there's a same bracket next to each other, remove one of them
                if len(open_brackets_position) != len(close_brackets_position):
                    filename = re.sub(f"({'|'.join(escaped_open_brackets)})\\1+", "\\1", filename)
                    filename = re.sub(f"({'|'.join(escaped_close_brackets)})\\1+", "\\1", filename)
                    open_brackets_position = []
                    close_brackets_position = []
                    for bracket in re.finditer(escaped_brackets, filename, re.IGNORECASE):
                        if bracket.group() in open_brackets:
                            open_brackets_position.append(bracket.start())
                        else:
                            close_brackets_position.append(bracket.start())

        # just to make sure that the brackets are sorted
        open_brackets_position.sort()
        close_brackets_position.sort()
        return filename, open_brackets_position, close_brackets_position

    def _tokenize_by_brackets(self, filename) -> str:
        filename, open_brackets_position, close_brackets_position = self._get_brackets_position(filename)

        cursor = 0
        last_pair_index = 0
        for open_bracket_index in open_brackets_position:
            if open_bracket_index < cursor:
                continue

            if open_bracket_index != cursor:
                # Found a token before the bracket
                self._tokenize_by_pre_identified(
                    filename[cursor:open_bracket_index],
                    enclosed=False
                )
                cursor = open_bracket_index
            for last_pair_index, close_bracket_index in enumerate(close_brackets_position[last_pair_index:]):
                if (close_bracket_index > open_bracket_index
                        and filename[close_bracket_index] == close_brackets[open_brackets.index(
                            filename[open_bracket_index])]):
                    break
            else:
                close_bracket_index = -1
            if close_bracket_index == -1:
                self._tokenize_by_pre_identified(
                    filename[cursor:],
                    enclosed=False
                )
                cursor = len(filename)
            else:
                self._add_token(
                    TokenCategory.BRACKET, filename[open_bracket_index], enclosed=True,
                    element=ElementCategory.BRACKET)
                self._tokenize_by_pre_identified(
                    filename[cursor + 1:close_bracket_index],
                    enclosed=True
                )
                self._add_token(
                    TokenCategory.BRACKET, filename[close_bracket_index], enclosed=True,
                    element=ElementCategory.BRACKET)
                cursor = close_bracket_index + 1
        else:
            self._tokenize_by_pre_identified(
                filename[cursor:],
                enclosed=False
            )
        return filename

    def _tokenize_by_pre_identified(self, text, enclosed) -> None:
        if not text:
            return

        pre_identified_tokens = self.keyword_manager.peek(text)

        last_token_end_pos = 0
        for token_begin_pos, token_end_pos, category in pre_identified_tokens:
            if last_token_end_pos != token_begin_pos:
                # Tokenize the text between the pre-identified tokens
                self._tokenize_by_delimiters(
                    text[last_token_end_pos:token_begin_pos], enclosed)
            self._add_token(TokenCategory.IDENTIFIER, text[token_begin_pos:token_end_pos], enclosed,
                            category)
            last_token_end_pos = token_end_pos

        if last_token_end_pos != len(text):
            # Tokenize the text after the pre-identified tokens (or all the text
            # if there was no pre-identified tokens)
            self._tokenize_by_delimiters(text[last_token_end_pos:], enclosed)

    def _tokenize_by_delimiters(self, text, enclosed) -> None:
        splited_text = re.split(self.pattern, text)
        new_tokens = Tokens(self.keyword_manager)  # Create a new Tokens object to store the new tokens
        for sub_text in splited_text:
            if sub_text:
                if sub_text in self.options['allowed_delimiters']:
                    new_tokens._add_token(TokenCategory.DELIMITER, sub_text, enclosed, ElementCategory.DELIMITER)
                else:
                    stripped = sub_text.strip(f" {parser_helper.DASHES}")
                    # e.g S01E01- > S01E01
                    match = re.match(f"(.*)({re.escape(stripped)})(.*)", sub_text)
                    if match:
                        if match.group(1):
                            # "-" in "- " or "-Flac"
                            new_tokens._add_token(TokenCategory.UNKNOWN, match.group(1), enclosed)
                        if match.group(2):
                            new_tokens._add_token(TokenCategory.UNKNOWN, match.group(2), enclosed)
                        if match.group(3):
                            # "-" in "S01E01-"
                            new_tokens._add_token(TokenCategory.UNKNOWN, match.group(3), enclosed)

        new_tokens._validate_delimiter_tokens()

        self.add_list(new_tokens.tokens)

    def tokenize(self) -> bool:
        self.filename = self._tokenize_by_brackets(self.filename)
        return not self.empty()



import logging

from aniparse import parser_helper, parser_number
from aniparse.element import ElementCategory
from aniparse.parser_number import ParserNumber
from aniparse.token import TokenCategory, TokenFlags, Token
from aniparse.tokenizer import Tokenizer

logger = logging.getLogger(__name__)


class Parser(Tokenizer, ParserNumber):

    def build_element(self, e_category: ElementCategory, token_begin=None, token_end=None) -> None:
        """
        Set the element category for all tokens between token_begin and token_end.

        :param e_category:
        :param token_begin:
        :param token_end:
        :return:
        """
        for token in self.get_list(begin=token_begin, end=token_end):
            if token.t_category == TokenCategory.UNKNOWN:
                self.set_token_element(token, TokenCategory.IDENTIFIER, e_category)
            elif token.t_category == TokenCategory.BRACKET:
                self.set_token_element(token, TokenCategory.IDENTIFIER, e_category)
            elif token.t_category == TokenCategory.DELIMITER:
                delimiter = token.content
                if self.options['keep_delimiters']:
                    self.set_token_element(token, TokenCategory.DELIMITER, e_category)
                elif token != token_begin and token != token_end:
                    if delimiter == ',' or delimiter == '&':
                        self.set_token_element(token, TokenCategory.IDENTIFIER, e_category)
                    else:
                        token.content = " "
                        self.set_token_element(token, TokenCategory.IDENTIFIER, e_category)
            else:
                self.set_token_element(token, TokenCategory.IDENTIFIER, e_category)

    def is_anime_season_keyword(self, token) -> bool:
        """
        Check if the token is a season keyword.
        :param token:
        :return:
        """
        # check for tre previous tokens
        previous_token = self.find_prev(token, TokenFlags.NOT_DELIMITER)
        if previous_token:
            number = parser_helper.get_number_from_ordinal(previous_token.content)
            if number:
                self.set_token_element(previous_token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_SEASON)
                previous_token.content = number
                return True

        # check for the next token
        next_token = self.find_next(token, TokenFlags.NOT_DELIMITER)
        if next_token:
            if next_token.content.isdigit():
                self.set_token_element(next_token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_SEASON)
                return True
        return False

    def parse(self) -> None:
        """
        Parse the tokens.
        First, the tokens are categorized. Then, find the indentifiable and searchable keyword.
        After that, look for isolated number for the year and video resolution. Finally, look for episode number.
        Once we have that information, we can determine the anime title, and episode title.
        Lastly, we can check for the release group and validate the elements.
        Any UNKNOWN token will be set to OTHER's category.
        :return:
        """
        if not self.tokenize():
            return None

        self.search_for_keywords()  # search for unique keywords for [year, video resolution, crc32, etc]

        self.search_for_isolated_number()  # search for isolated numbers for [season, anime year]

        self.search_for_episode_number()  # search for episode number. If it not exist, then it can't have episode

        first_number = None
        last_number = None
        if self.counter.get(ElementCategory.EPISODE_NUMBER, 0) == 2:
            first_number = self.find_next(None, element=ElementCategory.EPISODE_NUMBER)
            last_number = self.find_next(first_number, element=ElementCategory.EPISODE_NUMBER)
        elif self.counter.get(ElementCategory.VOLUME_NUMBER, 0) == 2:
            first_number = self.find_next(None, element=ElementCategory.VOLUME_NUMBER)
            last_number = self.find_next(first_number, element=ElementCategory.VOLUME_NUMBER)
        if first_number and last_number:
            tokens = list(self.get_list(flags=TokenFlags.UNKNOWN, begin=first_number, end=last_number))
            if len(tokens) == 1:
                if tokens[0].content in "&":
                    self.set_token_element(tokens[0], TokenCategory.INVALID, ElementCategory.RANGE_SEPARATOR)

        self.search_for_anime_title()

        self.search_for_episode_title()

        self.search_for_release_group()

        self.validate_elements()

    def search_for_anime_title(self) -> None:
        """
        Search for the anime title.
        First, we find not enclosed or unknown tokens. If we don't find any, then we
        check if `check_title_enclosed` is True. If it is, then we check for enclosed tokens.
        After that, we verify there is no Anime season, season prefix, episode number, volume number,
        or anime type before the anime title if `title_before_episode` is True.
        Finally, we can search the token_end for the anime title until it found an IDENTIFIER or BRACKET token.

        :return:
        """
        enclosed_title = False

        token_begin = self.find(TokenFlags.NOT_ENCLOSED | TokenFlags.UNKNOWN)
        if token_begin is None:
            if self.options['check_title_enclosed']:
                enclosed_title = True
                token_begin = self.get(0)
                skipped_previous_group = False

                while token_begin is not None:
                    token_begin = self.find_next(token_begin, TokenFlags.UNKNOWN | TokenFlags.IDENTIFIER)
                    if token_begin is None:
                        break

                    if (token_begin.e_category == ElementCategory.ANIME_SEASON_PREFIX
                            or token_begin.e_category == ElementCategory.ANIME_SEASON
                            or token_begin.e_category == ElementCategory.EPISODE_PREFIX
                            or token_begin.e_category == ElementCategory.EPISODE_NUMBER
                            or token_begin.e_category == ElementCategory.ANIME_TYPE):
                        if self.options['title_before_episode']:
                            break
                        else:
                            continue
                    # Ignore groups that are composed of non-Latin characters
                    if parser_helper.is_mostly_latin_string(token_begin.content):
                        if skipped_previous_group:
                            break  # Found it
                    # Get the first unknown token of the next group
                    token_begin = self.find_next(token_begin, TokenFlags.BRACKET)
                    skipped_previous_group = True

        if token_begin is None:
            return

        # if we see season prefix, number, or episode prefix, number, before token begin,
        # the stop searching for anime title
        for token in self.get_list(end=token_begin):
            if (self.options['title_before_episode']
                    and (token.e_category == ElementCategory.ANIME_SEASON_PREFIX
                         or token.e_category == ElementCategory.ANIME_SEASON
                         or token.e_category == ElementCategory.EPISODE_PREFIX
                         or token.e_category == ElementCategory.EPISODE_NUMBER
                         or token.e_category == ElementCategory.ANIME_TYPE)):
                return

        token_end = self.find_next(token_begin, TokenFlags.IDENTIFIER | (
            TokenFlags.BRACKET if enclosed_title else TokenFlags.NONE
        ))

        # If within the interval there's an open bracket without its matching
        # pair, move the upper endpoint back to the bracket
        if not enclosed_title:
            last_bracket = token_end
            bracket_open = False
            for token in self.get_list(TokenFlags.BRACKET, begin=token_begin, end=token_end):
                last_bracket = token
                bracket_open = not bracket_open
            if bracket_open:
                token_end = last_bracket

        # If the interval ends with an enclosed group (e.g. "Anime Title
        # [Fansub]"), move the upper endpoint back to the beginning of the
        # group. We ignore parentheses in order to keep certain groups (e.g.
        # "(TV)") intact.
        if not enclosed_title:
            token = self.find_prev(token_end, TokenFlags.NOT_DELIMITER)
            while token and token.t_category == TokenCategory.BRACKET and token.content != ")":
                token = self.find_prev(token, TokenFlags.BRACKET)
                if token:
                    token_end = token
                    token = self.find_prev(token_end, TokenFlags.NOT_DELIMITER)
            # Not sure why this is here, but it breaks the parser if not here
            token_end = self.find_prev(token_end, TokenFlags.VALID)
        potential_type = None
        anime_types = []
        for token in self.get_list(begin=token_begin, end=token_end):
            if (parser_helper.is_dash_character(token.content)
                    or token.t_category == TokenCategory.DELIMITER
                    or token.content.upper() == "CLEAN"):
                token_end = token
                continue
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(token.content))
            if keyword:
                element = keyword.e_category
                if element == ElementCategory.ANIME_TYPE:
                    anime_types.append(token)
                    if token.content.upper() in ["ED", "NCED", "ENDING",
                                                 "OP", "NCOP", "OPENING",
                                                 "PV", "PREVIEW"]:
                        clean_token = self.find_prev(token, TokenFlags.UNKNOWN)
                        if clean_token and clean_token.content.upper() == "CLEAN":
                            token_end = self.find_prev(clean_token)
                            token.content = f"{clean_token.content} {token.content}"
                            self.remove_token(clean_token)
                        self.set_token_element(clean_token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_TYPE)
                        for anime_type in anime_types:
                            self.set_token_element(anime_type, TokenCategory.IDENTIFIER, ElementCategory.ANIME_TYPE)
                        if potential_type:
                            token_end = potential_type
                        break
                    previous_token = self.find_prev(token, TokenFlags.NOT_DELIMITER)
                    if previous_token and parser_helper.is_dash_character(previous_token.content):
                        next_token = self.find_next(token, TokenFlags.NOT_DELIMITER)
                        if next_token:
                            if next_token.t_category != TokenCategory.UNKNOWN:
                                self.set_token_element(token, TokenCategory.IDENTIFIER,
                                                       ElementCategory.ANIME_TYPE)
                                break
                            if not potential_type:
                                potential_type = token_end
                        else:
                            break
                    if potential_type:
                        token_end = potential_type
            if potential_type != token_end:
                potential_type = None
                anime_types = []
            token_end = token
        if enclosed_title:
            token_end = self.find_prev(token_end, TokenFlags.VALID)
        self.build_element(ElementCategory.ANIME_TITLE, token_begin,
                           token_end)

    def search_for_episode_number(self) -> None:
        """
        Search for the episode number.
        First, we get all tokens that have number in it, Then we check for number pattern, buggy dash, equivalent number,
        seperated number, isolated number, and finally, last number as the last resort.
        :return:
        """
        tokens = []
        for token in self.get_list(TokenFlags.UNKNOWN):
            if parser_helper.find_number_in_string(token.content) is None:
                continue
            tokens.append(token)

        if not tokens:
            return

        if self.search_for_number_patterns(tokens):
            return

        tokens_original = tokens.copy()
        tokens = [token for token in tokens if token.content.isdigit()]

        if not tokens:
            # buggy dash, e.g. "Gunbuster-ep1"
            self.search_for_buggy_dash(tokens_original)
            return

        # e.g. "01 (176)", "29 (04)"
        if self.search_for_equivalent_numbers(tokens):
            return

        # e.g. " - 08"
        if self.search_for_seperated_numbers(tokens):
            return

        # e.g. "[12]", "(2006)"
        if self.search_for_isolated_numbers(tokens):
            return

        # consider using the last number as a last resort
        if self.search_for_last_number(tokens):
            return

        self.search_for_buggy_dash(tokens_original)

    def search_for_episode_title(self) -> None:
        """
        Search for the episode title.
        First, We find the UNKNOWN or NOT_ENCLOSED token. Then, we makesure there's no Audio/Video/Resolution/Source
        before the token begin. After that, we find the IDENTIFIER or ENCLOSED token for token end.
        :return:
        """
        token_end = None
        for _ in range(len(self.tokens)):

            # Find the first non-enclosed unknown token
            if token_end:
                token_begin = self.find_next(
                    token_end, TokenFlags.NOT_ENCLOSED | TokenFlags.UNKNOWN)
            else:
                token_begin = self.find(
                    TokenFlags.NOT_ENCLOSED | TokenFlags.UNKNOWN)

            if token_begin is None:
                return
            if token_begin.t_category == TokenCategory.IDENTIFIER:
                continue
            breaking = False
            for token in self.get_list(end=token_begin):
                # return if token before token_begin is Video Resolution, Audio Term, Video Term, etc
                if token.e_category in [
                    ElementCategory.AUDIO_TERM,
                    ElementCategory.DEVICE_COMPATIBILITY,
                    ElementCategory.FILE_CHECKSUM,
                    ElementCategory.FILE_EXTENSION,
                    ElementCategory.FILE_NAME,
                    ElementCategory.LANGUAGE,
                    ElementCategory.OTHER,
                    ElementCategory.RELEASE_INFORMATION,
                    ElementCategory.SOURCE,
                    ElementCategory.SUBTITLES,
                    ElementCategory.VIDEO_RESOLUTION,
                    ElementCategory.VIDEO_TERM
                ]:
                    breaking = True
                    break
            if breaking:
                break

            if self.counter.get(ElementCategory.EPISODE_NUMBER, 0) == 1:
                for token in self.get_list(begin=token_begin):
                    if token.e_category in [
                        ElementCategory.ANIME_SEASON,
                        ElementCategory.ANIME_SEASON_PREFIX,
                        ElementCategory.ANIME_TITLE,
                        ElementCategory.ANIME_TITLE_ALT,
                        ElementCategory.EPISODE_NUMBER,
                        ElementCategory.EPISODE_PREFIX,
                        ElementCategory.EPISODE_PART,
                        ElementCategory.EPISODE_TOTAL,
                    ]:
                        breaking = True
                        break

            # Continue until a bracket or identifier is found
            token_end = self.find_next(
                token_begin, TokenFlags.BRACKET | TokenFlags.IDENTIFIER)
            if breaking:
                continue

            # Ignore if it's only a dash
            if self.distance(token_begin, token_end) <= 2 and \
                    parser_helper.is_dash_character(token_begin.content):
                continue

            # If token end is a bracket, then we get the previous token to be
            # included in the element
            if token_end and token_end.t_category == TokenCategory.BRACKET:
                token_end = self.find_prev(token_end, TokenFlags.VALID)
            # Build episode title
            self.build_element(
                ElementCategory.EPISODE_TITLE, token_begin, token_end)
        return

    def search_for_isolated_number(self) -> None:
        """
        This function will search for isolated number for the year and video resolution.
        :return:
        """
        for token in self.get_list(TokenFlags.UNKNOWN):
            if (not token.content.isdigit()
                    or not self.is_token_isolated(token)):
                continue

            number = int(token.content)
            if parser_number.ANIME_YEAR_MIN <= number <= parser_number.ANIME_YEAR_MAX:
                self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_YEAR)
            elif number == 480 or number == 720 or number == 1080 or number == 2160:
                self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.VIDEO_RESOLUTION)

    def search_for_keywords(self) -> None:
        """
        This function will search for keywords about:
        - Anime season prefix
        - Episode number prefix
        - Volume number prefix
        - Release Version
        - File CRC32
        - Video resolution
        - Any identifiable keywords, keywords.options["default"]
        """
        prev = None
        for token in self.get_list(TokenFlags.UNKNOWN):
            word = token.content
            word = word.strip(' -')

            if not word:
                continue

            # skip if it is a number and not a CRC number
            if word.isdigit() and len(word) != 8:
                continue

            element = ElementCategory.UNKNOWN
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(word))
            if keyword:
                element = keyword.e_category

                if not ElementCategory.is_searchable(element):
                    continue
                if not keyword.options.searchable:
                    if element == ElementCategory.ANIME_TYPE:
                        previous_token = self.find_prev(token, TokenFlags.NOT_DELIMITER)
                        next_token = self.find_next(token, TokenFlags.NOT_DELIMITER)
                        if not next_token:
                            self.set_token_element(token, TokenCategory.IDENTIFIER, element)
                        elif parser_helper.is_dash_character(next_token.content):
                            self.set_token_element(token, TokenCategory.IDENTIFIER, element)
                        elif next_token and previous_token and previous_token.content == next_token.content:
                            self.set_token_element(token, TokenCategory.IDENTIFIER, element)
                    continue
                if not keyword.options.identifiable and not token.enclosed:
                    continue
                if not keyword.options.valid:
                    if not prev or not parser_helper.is_dash_character(prev.content):
                        continue

                if element == ElementCategory.ANIME_SEASON_PREFIX:
                    if self.is_anime_season_keyword(token):
                        self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_SEASON_PREFIX)
                    continue
                elif (element == ElementCategory.EPISODE_PREFIX
                      and keyword.options.valid):
                    if self.is_extent_keyword(token, ElementCategory.EPISODE_NUMBER, prefix=True):
                        self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.EPISODE_PREFIX)
                    continue
                elif element == ElementCategory.VOLUME_PREFIX:
                    if self.is_extent_keyword(token, ElementCategory.VOLUME_NUMBER, prefix=True):
                        self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.VOLUME_PREFIX)
                    continue
                elif element == ElementCategory.RELEASE_VERSION:
                    # this is probably a version number
                    token.content = word[0]
                    self.set_token_element(token, TokenCategory.INVALID, ElementCategory.UNKNOWN)
                    new_token = Token(word[1:], TokenCategory.IDENTIFIER, ElementCategory.RELEASE_VERSION,
                                      token.enclosed)
                    self.insert_after(token, new_token)
                    continue
                elif element == ElementCategory.AUDIO_TERM:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.AUDIO_TERM)
                    continue
                elif element == ElementCategory.DEVICE_COMPATIBILITY:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.DEVICE_COMPATIBILITY)
                    continue
                elif element == ElementCategory.LANGUAGE:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.LANGUAGE)
                    continue
                elif element == ElementCategory.OTHER:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.OTHER)
                    continue
                elif element == ElementCategory.RELEASE_INFORMATION:
                    if prev and prev.content.lower() in ["the"]:
                        continue
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.RELEASE_INFORMATION)
                    continue
                elif element == ElementCategory.SOURCE:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.SOURCE)
                    continue
                elif element == ElementCategory.SUBTITLES:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.SUBTITLES)
                    continue
                elif element == ElementCategory.VIDEO_TERM:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.VIDEO_TERM)
                    continue
                # last resort to avoid false positives
                if token.enclosed:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, element)
            else:
                # we could have multiple valid crc and resolution keywords.
                if parser_helper.is_crc32(word):
                    element = ElementCategory.FILE_CHECKSUM
                if parser_helper.is_resolution(word):
                    element = ElementCategory.VIDEO_RESOLUTION

            if element != ElementCategory.UNKNOWN:
                if keyword is None or keyword.options.identifiable:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, element)
            prev = token

    def search_for_release_group(self) -> None:
        """
        This function will search for the release group.
        It just look for the first UNKNOWN token that is enclosed.
        :return:
        """
        token_end = None
        for _ in range(len(self.tokens)):
            if token_end:
                token_begin = self.find_next(token_end, TokenFlags.UNKNOWN)
            else:
                token_begin = self.find(TokenFlags.UNKNOWN)
            if token_begin is None:
                return

            # Continue until a bracket or identifier is found
            token_end = self.find_next(token_begin, TokenFlags.BRACKET | TokenFlags.IDENTIFIER)
            if token_end is None:
                return

            if token_end.t_category != TokenCategory.BRACKET:
                continue

            # Ignore if it's not the first non-delimiter token in group
            previous_token = self.find_prev(token_begin, TokenFlags.NOT_DELIMITER)
            if previous_token and previous_token.t_category != TokenCategory.BRACKET:
                continue

            # Build release group, token end is a bracket, so we get the
            # previous token to be included in the element
            token_end = self.find_prev(token_end, TokenFlags.VALID)
            # get all tokens between token_begin and token_end
            tokens = self.get_list(begin=token_begin, end=token_end)
            content = "".join(token.content for token in tokens)
            replaced_token = Token(content, TokenCategory.IDENTIFIER, ElementCategory.RELEASE_GROUP, TokenFlags.NONE)
            index = self.get_index(token_begin)
            for token in tokens:
                self.remove_token(token)
            self.insert(index, replaced_token)
            return

    def validate_elements(self) -> None:
        """
        This function will validate the elements.
        It will check the episode number for the episode number alternatives.
        Then it will check for the episode title, to avoid anime type end up as episode title and some identifier token.
        And finally it check for anime title in case it ends with `the` and have anime type. (e.g. `the movie`)
        :return:
        """
        if self.counter.get(ElementCategory.EPISODE_NUMBER, 0) > 1:
            # get last episode number token
            token_episode = self.find_prev(None, flags=TokenFlags.IDENTIFIER, element=ElementCategory.EPISODE_NUMBER)
            token_start_episode = self.find_next(None, flags=TokenFlags.IDENTIFIER,
                                                 element=ElementCategory.EPISODE_NUMBER)
            token_start = self.find_next(token_start_episode, flags=TokenFlags.MASK_CATEGORIES)

            tokens = self.get_list(begin=token_start, end=token_episode)

            multi = False
            for token in tokens:
                if token.e_category == ElementCategory.RANGE_SEPARATOR:
                    multi = True
                    break
                if token.e_category == ElementCategory.EPISODE_NUMBER:
                    if float(token_start_episode.content) == float(token.content):
                        self.remove_token(token)
                        break

                    if self.options['eps_lower_than_alt']:
                        if float(token_start_episode.content) > float(token.content):
                            token = token_start_episode

                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.EPISODE_NUMBER_ALT)
                    break

            not_enclosed_tokens = [token_start_episode]
            enclosed_tokens = []
            for token in tokens:
                if token.e_category in [ElementCategory.EPISODE_NUMBER, ElementCategory.EPISODE_NUMBER_ALT]:
                    if token.enclosed:
                        enclosed_tokens.append(token)
                    else:
                        not_enclosed_tokens.append(token)

            if multi:
                if len(enclosed_tokens) == len(not_enclosed_tokens) == 2:

                    if self.options['eps_lower_than_alt']:
                        for idx, token in enumerate(enclosed_tokens):
                            if float(enclosed_tokens[idx].content) > float(not_enclosed_tokens[idx].content):
                                self.set_token_element(token, TokenCategory.IDENTIFIER,
                                                       ElementCategory.EPISODE_NUMBER_ALT)
                            else:
                                self.set_token_element(not_enclosed_tokens[idx], TokenCategory.IDENTIFIER,
                                                       ElementCategory.EPISODE_NUMBER_ALT)
                    else:
                        for token in enclosed_tokens:
                            self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.EPISODE_NUMBER_ALT)

        # Fill the unknown tokens with the keywords
        for token in self.get_list(TokenFlags.UNKNOWN):
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(token.content))
            if keyword:
                self.set_token_element(token, TokenCategory.IDENTIFIER, keyword.e_category)

        tokens = self.get_list(elements=ElementCategory.EPISODE_TITLE)
        real_token = []
        for token in tokens:
            if not token.content.strip(" " + parser_helper.DASHES):
                continue
            if token.t_category in [TokenCategory.DELIMITER, TokenCategory.BRACKET]:
                continue
            real_token.append(token)

        if len(real_token) == 1:
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(real_token[0].content))
            if keyword:
                self.set_token_element(real_token[0], TokenCategory.IDENTIFIER, keyword.e_category)
        elif len(real_token) == 2:
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(real_token[1].content))
            if keyword and keyword.e_category == ElementCategory.ANIME_TYPE:
                if (real_token[0].content.upper() == "CLEAN"
                        and real_token[1].content.upper() in ["ED", "NCED", "ENDING",
                                                              "OP", "NCOP", "OPENING"]):
                    self.set_token_element(real_token[0], TokenCategory.IDENTIFIER, keyword.e_category)
                    real_token[0].content += " " + real_token[1].content
                    self.remove_token(real_token[1])
        elif real_token:
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(real_token[-1].content))
            if keyword and keyword.e_category == ElementCategory.ANIME_TYPE:
                self.set_token_element(real_token[-1], TokenCategory.IDENTIFIER, keyword.e_category)

        # Check for unknown tokens between Season and Episode
        token_season = self.find_next(None, flags=TokenFlags.IDENTIFIER,
                                      element=[
                                          ElementCategory.ANIME_SEASON_PREFIX,
                                          ElementCategory.ANIME_SEASON,
                                      ])
        token_type = self.find_next(None, flags=TokenFlags.IDENTIFIER,
                                    element=[
                                        ElementCategory.ANIME_TYPE,
                                    ])
        token_number = self.find_next(None, flags=TokenFlags.IDENTIFIER,
                                      element=[
                                          ElementCategory.EPISODE_NUMBER,
                                      ])
        if token_season:
            token_end = token_type or token_number
            if token_end:
                actually_a_title = False
                for token in self.get_list(TokenFlags.UNKNOWN, token_season, token_end):
                    if parser_helper.is_dash_character(token.content) or not token.content:
                        continue
                    actually_a_title = True
                if actually_a_title:
                    # set token from end of anime title until season token to be anime title
                    anime_title = self.find_prev(token_season, flags=TokenFlags.IDENTIFIER)
                    for token in self.get_list(None, anime_title, token_end):
                        if token.e_category in [token_end.e_category]:
                            break
                        if token.e_category == ElementCategory.ANIME_SEASON:
                            token.content = str(int(token.content))
                        self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_TITLE)

        if token_type and token_number:
            actually_a_title = False
            for token in self.get_list(TokenFlags.UNKNOWN, token_type, token_number):
                if parser_helper.is_dash_character(token.content) or not token.content:
                    continue
                actually_a_title = True
            if actually_a_title:
                # set token from end of anime title until season token to be anime title
                anime_title = self.find_prev(token_type, flags=TokenFlags.IDENTIFIER)
                for token in self.get_list(None, anime_title, token_number):
                    if (token.e_category in [token_number.e_category, ElementCategory.ANIME_TYPE]
                            and token != token_type):
                        break
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_TITLE)

        # make sure there's no anime type in the episode title
        tokens = self.get_list(elements=ElementCategory.EPISODE_TITLE)
        is_other = False
        changed = []
        for token in tokens:
            keyword = self.keyword_manager.find(self.keyword_manager.normalize(token.content))
            if keyword:
                if keyword.options.identifiable:
                    changed.append([token, keyword])
                    is_other = True
                elif is_other:
                    self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.OTHER)
            else:
                changed = []

        for token, keyword in changed:
            self.set_token_element(token, TokenCategory.IDENTIFIER, keyword.e_category)

        # Anime type `Movie` are invalid if the last token is `the`
        token = self.find_prev(None, flags=TokenFlags.IDENTIFIER, element=ElementCategory.ANIME_TITLE)
        if token:
            if token.content.lower() == "the":
                anime_type = self.find_next(token, flags=TokenFlags.IDENTIFIER, element=ElementCategory.ANIME_TYPE)
                if anime_type:
                    tokens = self.get_list(begin=token, end=anime_type)
                    for token in tokens:
                        if token.t_category == TokenCategory.DELIMITER:
                            token.content = " "
                        self.set_token_element(token, TokenCategory.IDENTIFIER, ElementCategory.ANIME_TITLE)



import logging

from aniparse.element import ElementCategory
from aniparse.keyword import KeywordManager
from aniparse.aniparse import Aniparse

logger = logging.getLogger(__name__)

__all__ = ['parse']

default_options = {
    'allowed_delimiters': ' _.&+,|',
    'check_title_enclosed': True,
    'eps_lower_than_alt': True,
    'ignored_dash': True,
    'ignored_strings': [],
    'keep_delimiters': False,
    'max_extension_length': 4,
    'title_before_episode': True,

}


def parse(filename, options=None, ignore_errors=True, keyword_manager: KeywordManager = KeywordManager()):
    if options is None:
        options = {}

    # Add missing options
    for key, value in default_options.items():
        options.setdefault(key, value)
    try:
        parser = Aniparse(filename, options, keyword_manager)
        parser.parse()
        return parser.populate()
    except Exception as e:
        logger.error('Error parsing {}: {}'.format(filename, e))
        if ignore_errors:
            return {
                ElementCategory.FILE_NAME.value: filename
            }
        raise



from enum import Enum
from typing import List, Union

from aniparse.element import ElementCategory
from aniparse.keyword import KeywordManager


class TokenCategory(Enum):
    # Auto enumerate elements
    def __new__(cls):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    UNKNOWN = ()
    BRACKET = ()
    DELIMITER = ()
    IDENTIFIER = ()
    INVALID = ()


class TokenFlags(int):
    NONE = 0
    # Categories
    BRACKET = 1 << 0
    NOT_BRACKET = 1 << 1
    DELIMITER = 1 << 2
    NOT_DELIMITER = 1 << 3
    IDENTIFIER = 1 << 4
    NOT_IDENTIFIER = 1 << 5
    UNKNOWN = 1 << 6
    NOT_UNKNOWN = 1 << 7
    VALID = 1 << 8
    NOT_VALID = 1 << 9
    # Enclosed
    ENCLOSED = 1 << 10
    NOT_ENCLOSED = 1 << 11
    # Masks
    MASK_CATEGORIES = (
            BRACKET | NOT_BRACKET |
            DELIMITER | NOT_DELIMITER |
            IDENTIFIER | NOT_IDENTIFIER |
            UNKNOWN | NOT_UNKNOWN |
            VALID | NOT_VALID
    )
    MASK_ENCLOSED = ENCLOSED | NOT_ENCLOSED


class Token:
    def __init__(self, content: str = None, t_category: TokenCategory = TokenCategory.UNKNOWN,
                 e_category: ElementCategory = ElementCategory.UNKNOWN, enclosed: bool = False):
        self.t_category = t_category
        self.content = content
        self.enclosed = enclosed
        self.e_category = e_category

    def __repr__(self):
        return (f"Token(content = \"{self.content}\", category = {self.t_category}, "
                f"element = {self.e_category}), enclosed = {self.enclosed}")

    def check_elements(self, elements: ElementCategory = None):
        if elements:
            return self.e_category in elements
        return True

    # check if token is of a certain category
    def check_flags(self, flags: int = None):
        # Checking for flags is set for this token
        # e.g. if we're looking key with flags IDENTIFIER,
        # we compare the flags binary with the token's flags binary
        # and check if the IDENTIFIER bit is set
        if flags is None:
            return True

        # local function to check if a flag is equal
        def check_flag(flag: int):
            return (flags & flag) == flag

        # if MASK_ENCLOSED is set, check if enclosed is equal to the flag
        if flags & TokenFlags.MASK_ENCLOSED:
            if not check_flag(TokenFlags.ENCLOSED if self.enclosed else TokenFlags.NOT_ENCLOSED):
                return False

        # if MASK_CATEGORIES is set, check if category is equal to the flag
        if flags & TokenFlags.MASK_CATEGORIES:
            if check_flag(TokenFlags.BRACKET
                          if self.t_category == TokenCategory.BRACKET else TokenFlags.NOT_BRACKET):
                return True
            if check_flag(TokenFlags.DELIMITER
                          if self.t_category == TokenCategory.DELIMITER else TokenFlags.NOT_DELIMITER):
                return True
            if check_flag(TokenFlags.IDENTIFIER
                          if self.t_category == TokenCategory.IDENTIFIER else TokenFlags.NOT_IDENTIFIER):
                return True
            if check_flag(TokenFlags.UNKNOWN
                          if self.t_category == TokenCategory.UNKNOWN else TokenFlags.NOT_UNKNOWN):
                return True
            if check_flag(TokenFlags.VALID
                          if self.t_category != TokenCategory.INVALID else TokenFlags.NOT_VALID):
                return True
            return False
        return True


class Tokens:
    def __init__(self, keyword_manager: KeywordManager):
        self.tokens = []  # type: List[Token]
        self.counter = {}  # type: (ElementCategory, int)
        self.keyword_manager = keyword_manager

    def _add_token(self, category, content, enclosed, element=ElementCategory.UNKNOWN) -> Token:
        new_token = Token(content, category, element, enclosed)
        self.append(new_token)
        return new_token

    def _validate_delimiter_tokens(self) -> None:
        removed = {}

        def find_previous_valid_token(target_token) -> Token:
            return self.find_prev(target_token, TokenFlags.VALID)

        def find_next_valid_token(target_token) -> Token:
            return self.find_next(target_token, TokenFlags.VALID)

        def is_delimiter_token(target_token) -> bool:
            return target_token is not None and \
                   target_token.t_category == TokenCategory.DELIMITER

        def is_unknown_token(target_token) -> bool:
            return target_token is not None and \
                   target_token.t_category == TokenCategory.UNKNOWN

        def is_single_character_token(target_token) -> bool:
            return is_unknown_token(target_token) and len(target_token.content) == 1 and \
                   target_token.content != '-'

        def append_token_to(target_token, append_to) -> None:
            if target_token in removed:
                return
            if append_to in removed:
                append_to = removed[append_to]
            append_to.content += target_token.content
            # remove this token
            removed[target_token] = append_to

        for token in self.get_list():

            if token.t_category != TokenCategory.DELIMITER:
                continue

            delimiter = token.content
            if delimiter == ",":
                continue
            prev_token = find_previous_valid_token(token)
            next_token = find_next_valid_token(token)

            # Check for single-character tokens to prevent splitting group
            # names, keywords, episode number, etc.
            # come back later
            if delimiter != ' ' and delimiter != '_':
                if is_single_character_token(prev_token):
                    append_token_to(token, prev_token)
                    while is_unknown_token(next_token):

                        # 2nd "." in "R.O.D"
                        append_token_to(next_token, prev_token)
                        next_token = find_next_valid_token(next_token)
                        if is_delimiter_token(next_token) and \
                                next_token.content == delimiter:
                            # "." in "R.O."
                            append_token_to(next_token, prev_token)
                            next_token = find_next_valid_token(next_token)
                    continue
                if is_single_character_token(next_token):
                    # e.g. "." in "07.1", "." in "TrueHD5.1"
                    append_token_to(token, prev_token)
                    append_token_to(next_token, prev_token)
                    continue

            # Check for adjacent delimiters
            if is_unknown_token(prev_token) and is_delimiter_token(next_token):
                next_delimiter = next_token.content
                if delimiter != next_delimiter and delimiter != ',':
                    if next_delimiter == ' ' or next_delimiter == '_':
                        append_token_to(token, prev_token)  # e.g. "." in "Vol. ", "Ep. "

            elif is_delimiter_token(prev_token) and \
                    is_delimiter_token(next_token):
                prev_delimiter = prev_token.content
                next_delimiter = next_token.content
                if prev_delimiter == next_delimiter and \
                        prev_delimiter != delimiter:
                    self.set_token_element(token, TokenCategory.UNKNOWN, ElementCategory.UNKNOWN)  # e.g. "&" in "_&_"

            # Check for other special cases
            if delimiter == '&' or delimiter == '+':
                if is_unknown_token(prev_token) and \
                        is_unknown_token(next_token):
                    if prev_token.content.isdigit() and \
                            next_token.content.isdigit():
                        append_token_to(token, prev_token)
                        append_token_to(next_token, prev_token)  # e.g. "01+02"

        for token in removed:
            self.remove_token(token)

    def append(self, token: Token) -> None:
        self.tokens.append(token)
        self.counter[token.e_category] = self.counter.get(token.e_category, 0) + 1

    def add_list(self, tokens: List[Token]) -> None:
        for token in tokens:
            self.append(token)

    def distance(self, token1: Token = None, token2: Token = None) -> int:
        begin_index = 0 if token1 is None else self.get_index(token1)
        end_index = len(self.tokens) if token2 is None else self.get_index(token2)
        return abs(end_index - begin_index)

    def empty(self) -> bool:
        return len(self.tokens) == 0

    def find(self, flags: int) -> Token:
        return self.find_in_tokens(self.tokens, flags)

    @staticmethod
    def find_in_tokens(tokens: List[Token],
                       flags: int,
                       elements: Union[ElementCategory, List[ElementCategory]] = None
                       ) -> Union[Token, None]:
        # return the first token that matches the flags
        for token in tokens:
            if not token.check_elements(elements):
                continue
            if token.check_flags(flags):
                return token
        return None

    def find_next(self, token: Token = None, flags: int = None, element: Union[ElementCategory, List[ElementCategory]] = None) -> Union[Token, None]:
        tokens = self.tokens
        if token is not None:
            tokens = tokens[self.get_index(token) + 1:]
        return self.find_in_tokens(tokens, flags, element)

    def find_prev(self, token: Token = None, flags: int = None, element: Union[ElementCategory, List[ElementCategory]] = None) -> Union[Token, None]:
        tokens = self.tokens.copy()  # make copy of the list, but the tokens are still the same
        if token is None:
            tokens.reverse()
        else:
            tokens = tokens[max(self.get_index(token) - 1, 0)::-1]
        return self.find_in_tokens(tokens, flags, element)

    def get(self, index: int) -> Token:
        return self.tokens[index]

    def get_index(self, token: Token) -> int:
        return self.tokens.index(token)

    def get_list(self,
                 flags: int = None,
                 begin: Token = None,
                 end: Token = None,
                 elements: Union[ElementCategory, List[ElementCategory]] = None) -> List[Token]:

        begin_index = 0 if begin is None else self.get_index(begin)
        end_index = len(self.tokens) if end is None else self.get_index(end)
        tokens = self.tokens[begin_index:end_index + 1]

        if elements:
            tokens = (token for token in tokens if token.check_elements(elements))

        if flags:
            tokens = (t for t in tokens if t.check_flags(flags))
        return tokens

    def insert(self, index: int, token: Token) -> None:
        self.tokens.insert(index, token)
        self.counter[token.e_category] = self.counter.get(token.e_category, 0) + 1

    def insert_after(self, token: Token, new_token: Token) -> None:
        self.insert(self.get_index(token) + 1, new_token)

    def insert_before(self, token: Token, new_token: Token) -> None:
        self.insert(self.get_index(token), new_token)

    def insert_tokens(self, token: Token, word: str, category: TokenCategory,
                      element: ElementCategory) -> Token:
        new_token = Token(word, category, element, token.enclosed)
        self.insert_after(token,
                          new_token
                          )
        return new_token

    def is_token_isolated(self, token: Token) -> bool:
        previous_token = self.find_prev(token, TokenFlags.NOT_DELIMITER)
        if previous_token.t_category != TokenCategory.BRACKET:
            return False

        next_token = self.find_next(token, TokenFlags.NOT_DELIMITER)
        if next_token.t_category != TokenCategory.BRACKET:
            return False
        return True

    def remove_token(self, token: Token) -> None:
        self.counter[token.e_category] -= 1
        self.tokens.remove(token)

    def set_token_element(self, token: Token, t_category: TokenCategory, e_category: ElementCategory) -> None:
        self.counter[token.e_category] -= 1
        self.counter[e_category] = self.counter.get(e_category, 0) + 1
        token.t_category = t_category
        token.e_category = e_category



import re
from typing import Union

import unicodedata

DASHES = "-\u2010\u2011\u2012\u2013\u2014\u2015\u2212\uFE58\uFE63\uFF0D"
ESCAPED_DASHES = re.escape(DASHES)


def find_non_number_in_string(string) -> Union[int, None]:
    any_number = [char.isdigit() for char in string]
    if not all(any_number):
        return any_number.index(False)
    return None


def find_number_in_string(string) -> Union[int, None]:
    any_number = [char.isdigit() for char in string]
    if any(any_number):
        return any_number.index(True)
    return None


def get_number_from_ordinal(content) -> Union[str, None]:
    ordinals = {
        '1st': '1', 'First': '1',
        '2nd': '2', 'Second': '2',
        '3rd': '3', 'Third': '3',
        '4th': '4', 'Fourth': '4',
        '5th': '5', 'Fifth': '5',
        '6th': '6', 'Sixth': '6',
        '7th': '7', 'Seventh': '7',
        '8th': '8', 'Eighth': '8',
        '9th': '9', 'Ninth': '9'
    }
    return ordinals.get(str(content).lower(), None)


def is_crc32(string) -> bool:
    if len(string) != 8:
        return False
    try:
        int(string, 16)
    except ValueError:
        return False
    return True


def is_dash_character(string) -> bool:
    return string in DASHES


def is_latin_char(char):
    return is_latin_char.cache.setdefault(char, 'LATIN' in unicodedata.name(char))


is_latin_char.cache = {}


def is_mostly_latin_string(string: str):
    if len(string) == 0:
        return False
    latin_length = len([char for char in string if is_latin_char(char)])
    return latin_length / len(string) >= 0.5


def is_potential_number(text) -> bool:
    """
    1.0 -> False
    1 -> True
    1.1 -> True
    1.1.1 -> False
    01 -> True
    01.1 -> False
    """
    try:
        f_text = float(text)
        # Evangelion 1.0: You Are (Not) Alone
        if f_text.is_integer():
            return True
        if str(f_text) != text:
            return False
        return True
    except ValueError:
        return False


def is_resolution(string) -> bool:
    pattern = '\\d{3,4}([pP]|([xX\u00D7]\\d{3,4}))$'
    return bool(re.match(pattern, string))



from unittest import TestCase

import aniparse
from aniparse.element import ElementCategory
from aniparse.keyword import KeywordManager, option
from tests.fixtures.table import table


class TestAniparse(TestCase):

    def test_table(self):
        keyword = KeywordManager()
        keyword.add(ElementCategory.RELEASE_GROUP, option["default"], [
            'THORA'
        ])
        for index, expected in enumerate(table):
            filename = expected["file_name"]
            anime = aniparse.parse(filename, keyword_manager=keyword)
            self.assertEqual(expected, anime, 'on entry number %d' % index)



import enum


class ElementCategory(enum.Enum):
    ANIME_SEASON = 'anime_season'
    ANIME_SEASON_PREFIX = 'anime_season_prefix'
    ANIME_TITLE = 'anime_title'
    ANIME_TITLE_ALT = 'anime_title_alt'
    ANIME_TYPE = 'anime_type'
    ANIME_YEAR = 'anime_year'
    AUDIO_TERM = 'audio_term'
    DEVICE_COMPATIBILITY = 'device_compatibility'
    BRACKET = 'bracket'
    DELIMITER = 'delimiter'
    EPISODE_NUMBER = 'episode_number'
    EPISODE_PART = 'episode_part'
    EPISODE_NUMBER_ALT = 'episode_number_alt'
    EPISODE_PREFIX = 'episode_prefix'
    EPISODE_TITLE = 'episode_title'
    EPISODE_TOTAL = 'episode_total'
    FILE_CHECKSUM = 'file_checksum'
    FILE_EXTENSION = 'file_extension'
    FILE_NAME = 'file_name'
    LANGUAGE = 'language'
    OTHER = 'other'
    RANGE_SEPARATOR = 'range_separator'
    RELEASE_GROUP = 'release_group'
    RELEASE_INFORMATION = 'release_information'
    RELEASE_VERSION = 'release_version'
    SOURCE = 'source'
    SUBTITLES = 'subtitles'
    VIDEO_RESOLUTION = 'video_resolution'
    VIDEO_TERM = 'video_term'
    VOLUME_NUMBER = 'volume_number'
    VOLUME_PREFIX = 'volume_prefix'
    UNKNOWN = 'unknown'

    @classmethod
    def is_searchable(cls, category: 'ElementCategory') -> bool:
        searchable_categories = [
            cls.ANIME_SEASON_PREFIX,
            cls.ANIME_TYPE,
            cls.AUDIO_TERM,
            cls.DEVICE_COMPATIBILITY,
            cls.EPISODE_PREFIX,
            cls.FILE_CHECKSUM,
            cls.LANGUAGE,
            cls.OTHER,
            cls.RELEASE_GROUP,
            cls.RELEASE_INFORMATION,
            cls.RELEASE_VERSION,
            cls.SOURCE,
            cls.SUBTITLES,
            cls.VIDEO_RESOLUTION,
            cls.VIDEO_TERM,
            cls.VOLUME_PREFIX
        ]
        return category in searchable_categories

    def __contains__(self, item: 'ElementCategory') -> bool:
        return self.value == item.value



import logging
from typing import List, Union

import unicodedata

from aniparse.element import ElementCategory

__all__ = ['option', 'KeywordManager', 'BaseKeywordManager']

logger = logging.getLogger(__name__)


class KeywordOption:
    def __init__(self, identifiable=True, searchable=True, valid=True):
        self.identifiable = identifiable
        self.searchable = searchable
        self.valid = valid


class Keyword:
    def __init__(self, e_category: ElementCategory, options: KeywordOption):
        self.e_category = e_category
        self.options = options


option = {
    "default": KeywordOption(),
    "invalid": KeywordOption(valid=False),
    "unidentifiable": KeywordOption(identifiable=False),
    "unidentifiable_invalid": KeywordOption(identifiable=False, valid=False),
    "unidentifiable_unsearchable": KeywordOption(identifiable=False, searchable=False)

}


class BaseKeywordManager:
    def __init__(self):
        self._file_extensions = {}
        self._keys = {}

    def _get_keyword_container(self, category) -> dict:
        """
        File extensions are stored in a separate container
        """
        return self._file_extensions if category == ElementCategory.FILE_EXTENSION else self._keys

    def add(self, category: ElementCategory, options: KeywordOption, keywords: List[str]) -> None:
        keyword_container = self._get_keyword_container(category)
        for keyword in keywords:
            if not keyword:
                continue
            if keyword in keyword_container:
                continue

            keyword_container[keyword] = Keyword(category, options)
            logger.debug('Added keyword: {} with option {}'.format(keyword, options))

    def find(self, string: str, e_category: ElementCategory = ElementCategory.UNKNOWN) -> Union[Keyword, None]:
        """
        Find a keyword in the string based on the category

        :param string:
        :param e_category:
        :return:
        """
        logger.debug('Searching for keyword: {}'.format(string))
        keyword_container = self._get_keyword_container(e_category)
        if string not in keyword_container:
            return None
        keyword = keyword_container[string]
        if (e_category != ElementCategory.UNKNOWN
                and keyword.e_category != e_category):
            return None
        return keyword

    @staticmethod
    def normalize(string: str) -> str:
        """
        Remove accents and other special symbols
        NFKD: Normalization Form Compatibility Decomposition, (Normalize form with change the length)
        """
        nfkd = unicodedata.normalize('NFKD', string)
        without_accents = ''.join([c for c in nfkd if not unicodedata.combining(c)])
        return without_accents.upper()

    @staticmethod
    def peek(string: str) -> list:
        """
        this to handle the case where the string are seperated by a delimiter (space)
        e.g. "H264-Flac" -> ["H264", "Flac"]
        this early peek have many things to improve, but for now it's good enough
        """
        entries = [
            (ElementCategory.AUDIO_TERM, ['Dual Audio', 'Multi Audio']),
            (ElementCategory.VIDEO_TERM, ['H264', 'H.264', "10 bit", "10 bits", "8 bit", "8 bits"]),
            (ElementCategory.VIDEO_RESOLUTION, ['480p', '720p', '1080p', '2160p', '4K', "480i", "720i", "1080i"]),
            (ElementCategory.SUBTITLES, ['Multi Subs', 'Multiple Subtitle', 'Multiple Subtitles']),
            (ElementCategory.SOURCE, ['Blu-Ray'])
        ]

        pre_identified_tokens = []

        for category, keywords in entries:
            for keyword in keywords:
                keyword_begin_pos = string.lower().find(keyword.lower())
                if keyword_begin_pos != -1:  # Found the keyword in the string
                    keyword_end_pos = keyword_begin_pos + len(keyword)
                    pre_identified_tokens.append(
                        (keyword_begin_pos, keyword_end_pos, category))
        pre_identified_tokens = sorted(pre_identified_tokens,
                                       key=lambda x: (x[0], -x[1]))  # lowest begin pos, the highest end pos
        # check for overlapping tokens
        # e.g. (0, 3, "Sub") and (0, 4, "Subs")
        # if the first token is a substring of the second token, remove the first token
        last_token = (-1, -1, None)
        for token in pre_identified_tokens.copy():
            if token[0] == last_token[0]:
                pre_identified_tokens.remove(token)
            last_token = token

        return sorted(pre_identified_tokens)


class KeywordManager(BaseKeywordManager):
    """
    Default Keyword Manager
    """
    def __init__(self):
        super().__init__()

        self.add(ElementCategory.ANIME_SEASON_PREFIX, option["default"], [
            'SAISON', 'SEASON'
        ])

        self.add(ElementCategory.ANIME_SEASON_PREFIX, option["unidentifiable"], [
            'S'])

        self.add(ElementCategory.ANIME_TYPE, option["unidentifiable_unsearchable"], [
            'GEKIJOUBAN', 'MOVIE',
            'OAD', 'OAV', 'ONA', 'OVA',
            'SPECIAL', 'SPECIALS',
            'TV'
        ])

        self.add(ElementCategory.ANIME_TYPE,
                 option["unidentifiable_unsearchable"],
                 ['SP'])  # e.g. "Yumeiro Patissiere SP Professional"

        self.add(ElementCategory.ANIME_TYPE, option["unidentifiable_invalid"], [
            'ED', 'ENDING', 'NCED', "CLEAN ENDING",
            'NCOP', 'OP', 'OPENING', "CLEAN OPENING",
            'PREVIEW', 'PV'
        ])

        self.add(ElementCategory.AUDIO_TERM, option["default"], [
            # Audio channels
            '2.0CH', '2CH', '5.1', '5.1CH', 'DTS', 'DTS-ES', 'DTS5.1',
            'TRUEHD5.1',
            # Audio codec
            'AAC', 'AACX2', 'AACX3', 'AACX4', 'AC3', 'EAC3', 'E-AC-3',
            'FLAC', 'FLACX2', 'FLACX3', 'FLACX4', 'LOSSLESS', 'MP3', 'OGG',
            'VORBIS',
            # Audio language
            'DUALAUDIO', 'DUAL AUDIO', 'DUAL-AUDIO',
            'MULTIAUDIO', 'MULTI AUDIO', 'MULTI-AUDIO'
        ])

        self.add(ElementCategory.DEVICE_COMPATIBILITY, option["default"], [
            'IPAD3', 'IPHONE5', 'IPOD', 'PS3', 'XBOX', 'XBOX360'
        ])

        self.add(ElementCategory.DEVICE_COMPATIBILITY, option["unidentifiable"], [
            'ANDROID'
        ])

        self.add(ElementCategory.EPISODE_PREFIX, option["default"], [
            'EP', 'EP.', 'EPS', 'EPS.', 'EPISODE', 'EPISODE.', 'EPISODES',
            'CAPITULO', 'EPISODIO', 'FOLGE'
        ])

        self.add(ElementCategory.EPISODE_PREFIX, option["invalid"], [
            'E'
        ])  # single-letter episode keywords are not valid

        self.add(ElementCategory.FILE_EXTENSION, option["default"], [
            '3GP', 'AVI', 'DIVX', 'FLV', 'M2TS', 'MKV', 'MOV', 'MP4', 'MPG',
            'OGM', 'RM', 'RMVB', 'TS', 'WEBM', 'WMV'
        ])

        self.add(ElementCategory.FILE_EXTENSION, option["invalid"], [
            'AAC', 'AIFF', 'FLAC', 'M4A', 'MP3', 'MKA', 'OGG', 'WAV', 'WMA',
            '7Z', 'RAR', 'ZIP',
            'ASS', 'SRT'
        ])

        self.add(ElementCategory.LANGUAGE, option["default"], [
            'ENG', 'ENGLISH', 'ESPANOL', 'JAP', 'PT-BR', 'SPANISH', 'VOSTFR'
        ])

        self.add(ElementCategory.LANGUAGE, option["unidentifiable"], [
            'ESP', 'ITA'
        ])  # e.g. "Tokyo ESP", "Bokura ga Ita"

        self.add(ElementCategory.OTHER, option["default"], [
            'REMASTER', 'REMASTERED', 'UNCENSORED', 'UNCUT',
            'TS', 'VFR', 'WIDESCREEN', 'WS'
        ])

        self.add(ElementCategory.RELEASE_INFORMATION, option["default"], [
            'BATCH', 'COMPLETE', 'PATCH', 'REMUX'
        ])

        self.add(ElementCategory.RELEASE_INFORMATION, option["unidentifiable"], [
            'END', 'FINAL'
        ])  # e.g. "The End of Evangelion", "Final Approach"

        self.add(ElementCategory.RELEASE_VERSION, option["default"], [
            'V0', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6', 'V7', 'V8', 'V9', 'V10'
        ])

        self.add(ElementCategory.SOURCE, option["default"], [
            'BD', 'BDRIP', 'BLURAY', 'BLU-RAY',
            'DVD', 'DVD5', 'DVD9', 'DVD-R2J', 'DVDRIP', 'DVD-RIP',
            'R2DVD', 'R2J', 'R2JDVD', 'R2JDVDRIP',
            'HDTV', 'HDTVRIP', 'TVRIP', 'TV-RIP',
            'WEBCAST', 'WEBRIP'
        ])

        self.add(ElementCategory.SUBTITLES, option["default"], [
            'ASS', 'BIG5', 'DUB', 'DUBBED', 'HARDSUB', 'HARDSUBS', 'RAW',
            'SOFTSUB', 'SOFTSUBS', 'SUB', 'SUBBED', 'SUBTITLED',
            'MULTI SUBS', 'MULTI-SUBS', 'MULTISUB', 'MULTISUBS'
        ])

        self.add(ElementCategory.VIDEO_TERM, option["default"], [
            # Frame rate
            '23.976FPS', '24FPS', '29.97FPS', '30FPS', '60FPS', '120FPS',
            # Video codec
            '8BIT', '8-BIT', '10BIT', '10BITS', '10-BIT', '10-BITS',
            'HI10', 'HI10P', 'HI444', 'HI444P', 'HI444PP',
            'H264', 'H265', 'H.264', 'H.265', 'X264', 'X265', 'X.264',
            'AVC', 'HEVC', 'HEVC2', 'DIVX', 'DIVX5', 'DIVX6', 'XVID',
            # Video format
            'AVI', 'RMVB', 'WMV', 'WMV3', 'WMV9',
            # Video quality
            'HQ', 'LQ',
            # Video resolution
            'HD', 'SD'
        ])

        self.add(ElementCategory.VOLUME_PREFIX, option["default"], [
            'VOL', 'VOL.', 'VOLUME'
        ])
